import logging
import os

import genanki
from wordfreq import word_frequency
from pydub import AudioSegment
from pprint import pprint

from autoanki.DeckManager import template_decks


class DeckManager:
    """
    The class to make anki decks. Create the file using generate_deck_file()
    One of the most important concepts is the id. No matter what deck the word is in, it should have the same id so the
    same card in different decks can be remembered.
    This class makes extensive use of genanki, so understanding how genanki works
        is pretty significant for understanding this.
    """

    def __init__(
        self,
        debug_level,
        include_traditional=True,
        include_pinyin=True,
        include_pinyin_numbers=False,
        include_zhuyin=False,
        include_part_of_speech=True,
        include_audio=False,
        word_frequency_filter=None,
        hsk_filter=None,
    ):
        self.logger = logging.getLogger("autoanki.dckmngr")
        self.logger.setLevel(debug_level)
        self.deck = genanki.Deck(2020000110, "autoankiTesting")

        self.include_pinyin = include_pinyin
        self.include_pinyin_numbers = include_pinyin_numbers
        self.include_zhuyin = include_zhuyin
        self.include_traditional = include_traditional
        self.include_audio = include_audio
        self.include_part_of_speech = include_part_of_speech

        self.word_frequency_filter = word_frequency_filter
        self.hsk_filter = hsk_filter

        self.book_list = []

    def settings(
        self,
        include_traditional,
        include_part_of_speech,
        include_audio,
        include_pinyin,
        include_zhuyin,
        word_frequency_filter,
        hsk_filter,
    ):
        """
        Configures settings for what's in the deck, and how it looks
        """
        self.include_traditional = include_traditional
        self.include_part_of_speech = include_part_of_speech
        self.include_audio = include_audio
        self.include_pinyin = include_pinyin
        self.include_zhuyin = include_zhuyin
        self.word_frequency_filter = word_frequency_filter
        self.hsk_filter = hsk_filter

    def generate_deck_file(self, words: list[dict], deck_name: str, filename: str):
        """
        Generates a deck file from the database
        :param deck_name: The name of the deck to be created
        :param definitions_filename: The name of the file containing the definitions.
        :return:
        """

        # Number of valid cards that have been added to the deck
        num_of_valid_cards_added = 0

        # length = general_functions.file_len(definitions_filename)
        # self.deck.add_note()
        self.deck = genanki.Deck(2020000110, deck_name)
        WORD = 1
        WORD_TRADITIONAL = 2
        PINYIN = 3
        PINYIN_NUMBERS = 4
        ZHUYIN = 5
        JYUTPING = 6
        PART_OF_SPEECH = 7
        NUMBER_OF_STROKES = 8
        SUB_COMPONENTS = 9
        DEFINITION = 10
        FREQUENCY = 11
        HSK_LEVEL = 12
        TOCFL_LEVEL = 13
        AUDIO_PATH = 14
        IMAGE_PATH = 15
        CHARACTER_GRAPHIC = 16
        EXAMPLES = 17

        audio_file_list = []
        seen_audio_files = set(audio_file_list)

        for row in words:

            # Apply filters
            if self.word_frequency_filter and row[FREQUENCY]:
                if self.word_frequency_filter < row[FREQUENCY]:
                    continue
            if self.hsk_filter and row[HSK_LEVEL]:
                if self.hsk_filter > row[HSK_LEVEL]:
                    continue

            word = row[WORD]
            if not word:
                self.logger.error(f"Tried to add word with no word. Row: {row}")
                continue
            definition = "<br>" + row[DEFINITION]
            if not definition:
                self.logger.error(f"Tried to add word with no definition. Row: {row}")
                continue

            word_traditional = "Not found"
            if self.include_traditional:
                raw_word_traditional = row[WORD_TRADITIONAL]
                if raw_word_traditional:
                    if len(word) != len(raw_word_traditional):
                        self.logger.error(
                            f"{word} and {raw_word_traditional} should be the same length"
                        )
                    # Replace all matching characters with a dash
                    word_traditional = ""
                    for i in range(len(word)):
                        if word[i] == raw_word_traditional[i]:
                            word_traditional += "-"
                        else:
                            word_traditional += raw_word_traditional[i]

            # Pinyin
            pinyin = ""
            if row[PINYIN]:
                pinyin = row[PINYIN]
            if self.include_pinyin_numbers:
                pinyin = row[PINYIN_NUMBERS]
            if pinyin == "":
                pinyin = "Not found"

            # Zhuyin
            zhuyin = "<br>Not found"
            if self.include_zhuyin and row[ZHUYIN]:
                zhuyin = "<br>" + row[ZHUYIN]
            else:
                zhuyin = ""

            part_of_speech = row[PART_OF_SPEECH]
            if not part_of_speech:
                part_of_speech = "Not found"

            # Audio
            audio = ""
            if self.include_audio and len(word) < 5:

                # Get the filename
                audio_name = str(row[PINYIN_NUMBERS])
                audio_name.replace("ü", "u")
                audio_name = audio_name.replace(" ", "")
                audio_name = audio_name.replace("'", "")

                audio_file = audio_name + ".mp3"

                # If it's not there, create it from spliced files
                path_to_audio_file = os.path.join(
                    os.path.dirname(__file__), "mandarin_sounds", audio_file
                )
                path_to_spliced_audio_file = os.path.join(
                    os.path.dirname(__file__), "mandarin_sounds", "spliced", audio_file
                )
                self.logger.debug(f"Audio file path: {path_to_audio_file}")
                if os.path.exists(path_to_audio_file):
                    audio_path = path_to_audio_file
                elif os.path.exists(path_to_spliced_audio_file):
                    audio_path = path_to_spliced_audio_file
                else:
                    audio_path = create_chinese_audio_file(audio_name)

                audio = f"[sound:{audio_file}]"

                # Look for the audio file
                if audio_path not in seen_audio_files:
                    seen_audio_files.add(audio_path)
                    audio_file_list.append(audio_path)

            card_tags = ["autoanki"]

            note = genanki.Note(
                model=template_decks.CHINESE_CARD_MODEL,
                tags=card_tags,
                fields=[
                    word,
                    word_traditional,
                    pinyin,
                    audio,
                    zhuyin,
                    part_of_speech,
                    definition,
                ],
                # sort_field can be used to sort when the cards appear.
                # By default they are shown in the order they are addeed, so this is not currently used
                sort_field=1,
            )
            self.deck.add_note(note)
            num_of_valid_cards_added += 1

        if not filename.endswith(".apkg"):
            filename += ".apkg"

        # pprint("Audio files:")
        # pprint(audio_file_list)
        my_package = genanki.Package(self.deck)
        my_package.media_files = audio_file_list
        my_package.write_to_file(filename)

        self.logger.info(
            "Deck "
            + deck_name
            + " created with "
            + str(num_of_valid_cards_added)
            + " cards"
        )
        return filename


def create_chinese_audio_file(audio_name: str) -> str:
    """Create a chinese audio file by splicing other files together
    This function assumes the folder with all of the base mp3s is in the right spot
    Return: path to the audio file
    """

    # print("Audio: " + audio_name)
    audio_name = audio_name.replace(" ", "")
    audio_path = os.path.join(
        os.path.dirname(__file__), "mandarin_sounds", "spliced", audio_name + ".mp3"
    )

    file_list = []
    current_pinyin = ""
    for char in audio_name:
        current_pinyin += char
        if char.isdigit():
            current_pinyin = current_pinyin.replace("ü", "u")
            current_pinyin = current_pinyin.strip()
            file_list.append(current_pinyin)
            current_pinyin = ""

    # print(file_list)

    # If the pre-made list doesnt have a 5th tone file, just use first tone
    for i, file in enumerate(file_list):
        full_path = os.path.join(
            os.path.dirname(__file__), "mandarin_sounds", file_list[0] + ".mp3"
        )
        if not os.path.exists(full_path):
            file_list[i] = file.replace("5", "1")

    try:
        # print("Grabbing files...")
        file_path = os.path.join(
            os.path.dirname(__file__), "mandarin_sounds", file_list[0] + ".mp3"
        )
        # print(file_path)
        total_sound = AudioSegment.from_mp3(file_path)
        for file in file_list[1:]:
            file_to_grab = os.path.join(
                os.path.dirname(__file__), "mandarin_sounds", file + ".mp3"
            )
            # print(file_to_grab)
            sound = AudioSegment.from_mp3(file_to_grab)
            total_sound += sound

        # Writing mp3 files is a one liner
        total_sound.export(audio_path, format="mp3")

        print("This will be saved to: " + audio_path)
        print()

        return audio_path

    except Exception as e:
        print(f"Error with {file_list}")
        print(e)
        return ""



from abc import ABC, abstractmethod


class Dictionary(ABC):
    @abstractmethod
    def __init__(self, debug_level=20):
        pass

    @abstractmethod
    def find_word(self, word: str) -> None | dict[str, str]:
        pass

    @abstractmethod
    def size(self) -> int:
        pass



import logging
import pandas as pd
import pdfplumber
import os
import ebooklib
from ebooklib import epub
from bs4 import BeautifulSoup

from autoanki.BookCleaner import BookCleaner
from autoanki.DatabaseManager import ChineseDatabaseManager
from autoanki.DatabaseManager.DatabaseManager import DatabaseManager
from autoanki.Dictionary import CEDictionary
from autoanki.DeckManager import DeckManager
from autoanki.Dictionary.Dictionary import Dictionary
from autoanki.Tokenizer import ChineseTokenizer

import time

BLACK = "\u001b[30m"
RED = "\u001b[31m"
GREEN = "\u001b[32m"
YELLOW = "\u001b[33m"
BLUE = "\u001b[34m"
MAGENTA = "\u001b[35m"
CYAN = "\u001b[36m"
WHITE = "\u001b[37m"
RESET = "\u001b[0m"
logging.basicConfig(
    # filename='HISTORY.log',
    level=logging.WARNING,
    format=f"{GREEN}%(asctime)s{RESET} {RED}%(levelname)8s{RESET} {YELLOW}%(name)-16s{RESET}: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class LanguageResources:
    database_manager: DatabaseManager
    dictionary: Dictionary
    deck_manager: DeckManager


# class Chinese(LanguageResources):
# database_manager: ChineseDatabaseManager
# dictionary_manager: CEDictionary
# deck_manager: DeckManager

supported_languages = {"zh": "Yes"}


class AutoAnki:
    def __init__(
        self,
        language: str,
        database_filepath: str = "",
        debug_level=20,
        force=False,
        dictionary=None,
    ):
        """
        Creates an instance of autoanki.
        This creates a book cleaner, database connection, dictioary connection, and deck maker
        Args:
            `language`: The language to load, as per its ISO 639-1 two-letter code (en, zh, fr...)
            `database_filepath`: The filepath for the database. If none specified a new one will be created
            `logging_level`: between 0 (DEBUG) and 50(CRITICAL)
            `force`: Skip conformations for cleaning large numbers of files
        """
        self.logger = logging.getLogger("autoanki")
        self.logger.setLevel(debug_level)
        self.logger.debug(f"Autoanki logger active")

        if len(language) != 2:
            self.logger.warning(
                f"Incorrect language input: [{language}]. Please use it's 2-letter code"
            )
            return

        if not supported_languages.get(language):
            self.logger.warning(f"Unsupported language: [{language}]")
            return

        self.logger.info(
            f"===== {GREEN}Starting to load config: [{language}] {RESET}====="
        )
        total_start = time.time()
        self.force = force

        # self.book_cleaner = BookCleaner(debug_level, self.force)

        start = time.time()
        self.logger.info("Loading dictionary...")
        if dictionary:
            self.logger.info("Using custom dictionary")
            self.dictionary = dictionary
        else:
            self.dictionary = CEDictionary(debug_level)
        end = time.time()
        self.logger.info(f"Done in {end - start:0.4f} seconds")

        self.database_filepath = database_filepath
        if database_filepath == "":
            self.logger.info("No database specified. Creating a new one...")
            timestr = time.strftime("%Y%m%d_%H%M%S")
            self.database_filepath = str("autoanki_" + timestr + ".db")
            ChineseDatabaseManager.create_database(self.database_filepath)
        else:
            if not ChineseDatabaseManager.is_database(self.database_filepath):
                self.logger.info("Creating database...")
                ChineseDatabaseManager.create_database(self.database_filepath)
                self.logger.info("Done creating database.")

        start = time.time()
        self.logger.info("Connecting to database...")
        self.database_manager = ChineseDatabaseManager(
            self.database_filepath, debug_level, dictionary=self.dictionary
        )
        end = time.time()
        self.logger.info(f"Done in {end - start:0.4f} seconds")

        start = time.time()
        self.logger.info("Connecting to DeckManager...")
        self.deck_manager = DeckManager(debug_level)
        end = time.time()
        self.logger.info(f"Done in {end - start:0.4f} seconds")

        total_end = time.time()
        self.logger.info(
            f"===== {GREEN}Done loading profile in {total_end - total_start:0.4f} seconds {RESET}====="
        )

    def add_book_from_string(self, contents: str, book_name: str = "Book Name"):
        """
        Add a book as a string to the database
        Args:
            `contents`: the contents of the book
            `book_name`: The name of the book being added e.g. "Lost Prince"
        """
        self.logger.debug(f"autoanki: Adding book [{book_name}] from string")
        if not contents:
            self.logger.info(f"No contents supplied")
            return

        # TODO Handle pdfs in file
        # TODO Restructure so each file can be passed to database_manager and still get added to the same book

        # Add the book to the database
        if not self.database_manager.add_book_from_string(contents, book_name):
            self.logger.warning("Unable to add [" + book_name + "] to database.")
            return

        self.logger.info("autoanki: Added book from string.")

    def add_book_from_file(self, filepath: str, book_name: str = "Book Name"):
        """Add a file to the database
        Args:
            `filepath`: path to the directory that contains the files to add
            `book_name`: The name of the book being added e.g. "Lost Prince"
        """
        self.logger.debug(
            f"autoanki: Adding book [{book_name}] from file: [{filepath}]"
        )
        if not filepath:
            self.logger.info(f"No filepath supplied")
            return
        # pip3 install pdfplumber

        # Handle pdf
        extension = os.path.splitext(filepath)[1]

        if extension == ".pdf":
            self.logger.info(f"PDF detected")
            # for every page
            print(filepath)
            print(os.path.exists(filepath))

            with pdfplumber.open(filepath) as pdf:
                print(pdf)
                for pages in pdf.pages:
                    print(pages.pages)
                    print(pages.extract_text())
                    if not self.database_manager.add_book_from_string(
                        pages.extract_text(), book_name
                    ):
                        self.logger.warning(
                            "Unable to add [" + book_name + "] to database."
                        )
                        return
        elif extension == ".epub":
            book = epub.read_epub(filepath)
            # Get the chapters
            chapters = ""
            for html_element in book.get_items_of_type(ebooklib.ITEM_DOCUMENT):
                self.logger.info(html_element)
                chapters += str(html_element.get_content(), "utf-8")
            blacklist = [
                "[document]",
                "noscript",
                "header",
                "html",
                "meta",
                "head",
                "input",
                "script",
            ]
            # These come out as html, so convert to text
            contents = ""
            soup = BeautifulSoup(chapters, "html.parser")
            text = soup.get_text()

            text = soup.find_all(text=True)
            for t in text:
                if t.parent.name not in blacklist:
                    contents += "{} ".format(t)
            if not self.database_manager.add_book_from_string(contents, book_name):
                self.logger.warning("Unable to add [" + book_name + "] to database.")
                return
        elif extension == ".txt":
            self.logger.info(f"txt detected")
            # Add the book to the database
            if not self.database_manager.add_book_from_file(filepath, book_name):
                self.logger.warning("Unable to add [" + book_name + "] to database.")
                return

        self.logger.info("autoanki: Added [" + filepath + "].")

    def add_book_from_folder(self, directory: str, book_name: str = "Book Name"):
        """
        Add a directory full of files to the database
        Args:
            `filepath`: path to the directory that contains the files to add
            `book_name`: The name of the book being added e.g. "Lost Prince"
        """
        self.logger.debug(
            f"autoanki: Adding book [{book_name}] from directory: [{directory}]"
        )
        if not directory:
            self.logger.info(f"No directory supplied")
            return

        # Add the book to the database
        if not self.database_manager.add_book_from_folder(directory, book_name):
            self.logger.warning("Unable to add [" + book_name + "] to database.")
            return

        self.logger.info("autoanki: Added [" + directory + "].")

    def add_book_from_pleco(self, filepath: str, book_name: str):
        """Reads the contents of a Pleco export file (txt, not xml)
        Args:
            `filepath`: path to the directory that contains the files to add
            `book_name`: The name of the book being added e.g. "Lost Prince"
        """
        self.logger.debug(
            f"autoanki: Adding book [{book_name}] from pleco: [{filepath}]"
        )
        if not filepath:
            self.logger.info(f"No filepath supplied")
            return

        with open(filepath, "r") as file:
            contents = file.read()
            for line in file.readlines():
                split_line = line.split(" ")
                if split_line:
                    contents += split_line[0]

        contents = ""
        self.logger.debug(f"Done reading Pleco. len: {len(contents)}")
        self.database_manager.add_book_from_string(contents, book_name)

    def complete_unfinished_definitions(self):
        """
        autoanki contains an internal definitions table that is scraped from the internet. As words are added to
        autoanki, their definitions must be found.
        This function finds definitions and adds them to the table
        """
        start = time.time()
        self.logger.info("Checking for records...")
        response_rows = self.database_manager.unfinished_definitions()
        if len(response_rows) == 0:
            self.logger.info("No new rows to complete in dictionary table")
            return

        self.logger.info(
            "Adding " + str(len(response_rows)) + " rows to dictionary table"
        )
        self.tokenizer = ChineseTokenizer(dictionary=self.dictionary)
        for row in response_rows:
            word = str(row[0])

            # self.logger.debug(f"Finding: [{word}]")
            # self.logger.debug("Trying local dictionary...")
            params = self.dictionary.find_word(word)
            # self.logger.info(params)
            if params:
                # self.logger.debug(f"✅Found: [{params[8]}]")
                self.database_manager.update_definition(params)
                continue

            self.logger.info(f"❌Could not find: [{word}]")
        end = time.time()
        self.logger.info(
            f"Finished collecting definitions in {end - start:0.4f} seconds"
        )

    def deck_settings(
        self,
        include_traditional=True,
        include_part_of_speech=True,
        include_audio=False,
        include_pinyin=True,
        include_zhuyin=False,
        hsk_filter=None,
        word_frequency_filter=None,
    ):
        """Configures settings for what's in the deck, and how it looks"

        `word_frequency_filter`: Float between 0 and 1. 1 being every word is included, 0 being none are included
        """
        self.deck_manager.settings(
            include_traditional=include_traditional,
            include_pinyin=include_pinyin,
            include_zhuyin=include_zhuyin,
            include_part_of_speech=include_part_of_speech,
            include_audio=include_audio,
            word_frequency_filter=word_frequency_filter,
            hsk_filter=hsk_filter,
        )

    def print_database_info(self):
        self.database_manager.print_info()

    @staticmethod
    def is_database(db_path):
        return ChineseDatabaseManager.is_database(db_path)

    @staticmethod
    def create_database(db_path: str):
        ChineseDatabaseManager.create_database(db_path)

    def create_deck(self, deck_name: str, filepath: str):
        """
        Creates a deck file in the directory of the main file.
        `deck_name` The name that will show up in Anki
        `filepath` Path to the file
        :return:
        """

        self.logger.info("Generating deck file [" + deck_name + ".apk]")
        words = self.database_manager.get_all_completed_definitions()

        deck_path = self.deck_manager.generate_deck_file(words, deck_name, filepath)
        if deck_path is None:
            self.logger.warning(
                "Was not able to create deck file for [", deck_name, "]"
            )
        else:
            self.logger.info("Generated deck file [" + deck_path + "]")

    def save_dictionary_as_csv(self, filepath: str):
        self.logger.info("Saving to csv...")
        all = self.database_manager.get_all_definitions()
        # pprint(all)

        df = pd.DataFrame(all)
        headers = self.database_manager.get_columns()
        df.columns = headers
        df.to_csv(filepath)
        self.logger.info("Done saving to csv...")

    @property
    def book_list(self):
        """
        Get a list of the books in the database
        :return: List of book names
        """
        return self.database_manager.books

    @book_list.setter
    def book_list(self, _):
        pass

    @property
    def unfinished_entries(self):
        return self.database_manager.unfinished_definitions()

    @unfinished_entries.setter
    def unfinished_entries(self, _):
        pass
